-- we need to use numeric IDs or issues with the achievement frame sub-category indenting occur
local FIRST_CATEGORY = 1337
local CURRENT_CATEGORY

local localfuncs = (select(2,...))

-- GLOBALS: FirstLootEver_Items, AlertFrame_AnimateIn, AlertFrame_FixAnchors, AchievementFrame_LoadUI, AchievementFrameCategories_Update, AchievementAlertFrame_GetAlertFrame, AchievementFrame_OnEvent, WatchFrame_OnEvent, AchievementFrameAchievements_OnEvent, AchievementFrame_ForceUpdate, AchievementFrameCategories_GetCategoryList, AlertFrame_OnEvent
-- GLOBALS: ACHIEVEMENTUI_CATEGORIES
-- GLOBALS: AchievementFrame, WatchFrame, AchievementFrameAchievements, AlertFrame
-- GLOBALS: GetCategoryInfo, GetCategoryNumAchievements, GetAchievementInfo, IsTrackedAchievement, GetAchievementNumCriteria, GetPreviousAchievement, GetNextAchievement, AddTrackedAchievement, RemoveTrackedAchievement, GetAchievementCriteriaInfo, GetAchievementCategory, GetAchievementLink, GetTrackedAchievements, GetNumTrackedAchievements
-- GLOBALS: SLASH_FIRSTLOOTEVER1
-- GLOBALS: _
local _G = _G
local unpack, pairs, ipairs, next, tinsert, tremove, print, type, tonumber, select, date, time, setmetatable, math, bit, table, tContains =
	  unpack, pairs, ipairs, next, tinsert, tremove, print, type, tonumber, select, date, time, setmetatable, math, bit, table, tContains
local GetRealmName, GetContainerNumSlots, GetContainerItemID, CreateFrame, GetPlayerInfoByGUID, UnitGUID, SetDesaturation =
	  GetRealmName, GetContainerNumSlots, GetContainerItemID, CreateFrame, GetPlayerInfoByGUID, UnitGUID, SetDesaturation
local BACKPACK_CONTAINER, NUM_BAG_SLOTS, ACHIEVEMENT_UNLOCKED, ACHIEVEMENT_TOOLTIP_COMPLETE, ACHIEVEMENT_TOOLTIP_IN_PROGRESS =
	  BACKPACK_CONTAINER, NUM_BAG_SLOTS, ACHIEVEMENT_UNLOCKED, ACHIEVEMENT_TOOLTIP_COMPLETE, ACHIEVEMENT_TOOLTIP_IN_PROGRESS
local GameTooltip =
	  GameTooltip

local dummyFunc = function() end
local gii = GetItemInfo
local iteminfo_cache = {}
local function GetItemInfo(i)
	if iteminfo_cache[i] then return unpack(iteminfo_cache[i]) end
	local retVals = {gii(i)}
	if not retVals[1] then
		-- automatically retry if the client might just not have cached the item yet
		retVals = {gii(i)}
	end
	if not retVals[1] then
		return nil
	end
	iteminfo_cache[i] = retVals
	return unpack(retVals)
end
localfuncs.gii = GetItemInfo

local giqc = GetItemQualityColor
local qualcolor_cache = {}
local function GetItemQualityColor(q)
	if qualcolor_cache[q] then return unpack(qualcolor_cache[q]) end
	local retvals = {giqc(q)}
	if not retvals[1] then return nil end
	qualcolor_cache[q] = retvals
	return unpack(retvals)
end
localfuncs.giqc=GetItemQualityColor

local addon = CreateFrame("Frame", "FirstLootEver")
--[[local itemGroups = {}
local reverseGroups = {}
function addon:RegisterItem(itemID, itemGroup)
	local g = itemGroups
	for i, k in ipairs(itemGroup) do
		if i > 3 then break end -- maximum of 3 sub-category levels (tab, sub-tab, label)
		g[k] = g[k] or {}
		g = g[k]
	end
	tinsert(g,itemID)
	local tableValue
	for i,k in ipairs(itemGroup) do -- can't use table.concat due to the 3 sub limit - table.concat doesn't allow any limiting of key number
		if i > 3 then break end
		tableValue = (tableValue and tableValue.."." or "")..k
	end
	reverseGroups[itemID] = tableValue
end--]]

local itemList = {}
localfuncs.itemlist= function() return itemList end
local filterOptions = {}
local filterLabels = {
	[1] = "Slot",
	[2] = "Quality",
	[3] = "Type",
	[4] = "Expansion",
	[5] = "Weapon type",
}
function addon:RegisterItem(itemID, itemFilterMatches)
	local num, filter = next(itemFilterMatches, nil)
	local lastNum
	while (num and filter) do
		if filter == "" then
			itemFilterMatches[num] = nil
			num, filter = next(itemFilterMatches,lastNum)
		else
			filterOptions[num] = filterOptions[num] or {}
			if not tContains(filterOptions[num], filter) then
				tinsert(filterOptions[num], filter)
			end
			lastNum=num
			num,filter = next(itemFilterMatches,num)
		end
	end
	itemList[itemID] = itemFilterMatches
	GetItemInfo(itemID) -- just to fill the cache for later
end

function addon:RegenerateItemCategoryListings()
end

local achievementGroups = {}
local reverseAchievements = {}
local blizzAchievements = {[""]=FIRST_CATEGORY}
local reverseBlizzAchievements = {}
local categoryListingCache = {}
function addon:RegisterAchievement(id, name, desc, icon, checkFunc, criteria, category, prevAch, nextAch)
	if not achievementGroups[category] then
		achievementGroups[category] = {}
		if not (category == "") then
			CURRENT_CATEGORY = CURRENT_CATEGORY or (FIRST_CATEGORY+1)
			reverseBlizzAchievements[CURRENT_CATEGORY] = category
			blizzAchievements[category] = CURRENT_CATEGORY
			CURRENT_CATEGORY = CURRENT_CATEGORY + 1
		end
	end
	if prevAch then
		local prevData = reverseAchievements[prevAch]
		tinsert(achievementGroups[prevData.category][prevData.idInCat], {id=id, icon=icon, checkFunc=checkFunc, criteria=criteria, desc=desc, name=name, prevAch=prevAch, nextAch=nextAch})
		reverseAchievements[id] = {category=prevData.category, idInCat=prevData.idInCat, icon=icon, checkFunc=checkFunc, criteria=criteria, desc=desc, name=name, prevAch=prevAch, nextAch=nextAch}
	else
		tinsert(achievementGroups[category], {{id=id, icon=icon, checkFunc=checkFunc, criteria=criteria, desc=desc, name=name, prevAch=prevAch, nextAch=nextAch}})
		reverseAchievements[id] = {category=category, idInCat=#achievementGroups[category], icon=icon, checkFunc=checkFunc, criteria=criteria, desc=desc, name=name, prevAch=prevAch, nextAch=nextAch}
	end
end

function addon:RegenerateAchievementCategoryListings()
	if not self.isLoaded then
		self.scheduledGenerate = true
		return
	end
	for category, data in pairs(achievementGroups) do
		categoryListingCache[category] = categoryListingCache[category] and table.wipe(categoryListingCache[category]) or {}
		local cache = categoryListingCache[category]
		local completed=0
		-- get last completed
		for _, progTable in ipairs(data) do
			for i=#progTable, 1, -1 do
				local avData = progTable[i]
				if self:HasAchievement(avData.id) then
					tinsert(cache, avData.id)
					completed = completed+1
					break
				end
			end
		end
		-- loop again, get first incomplete
		for _, progTable in ipairs(data) do
			for _, avData in ipairs(progTable) do
				if (not self:HasAchievement(avData.id)) then
					tinsert(cache, avData.id)
					break
				end
			end
		end
		cache.numCompleted = completed
	end
	if not categoryListingCache[""] then
		categoryListingCache[""] = {numCompleted = 0}
	end
	if AchievementFrame.selectedTab == 1 then
		AchievementFrameCategories_GetCategoryList(ACHIEVEMENTUI_CATEGORIES)
	end
	AchievementFrameCategories_Update()
end

function addon:HasAchievement(id)
	if self.savedVar.achievements and self.savedVar.achievements[id] then
		return true
	end
end

function addon:AwardAchievement(id)
	self.savedVar.achievements = self.savedVar.achievements or {}
	self.savedVar.achievements[id] = time()
	self:ShowAchievementPopup(id)
	self:RegenerateAchievementCategoryListings()
	AchievementFrame_ForceUpdate()
end

function addon:ShowAchievementPopup(id)
	AlertFrame_OnEvent(AlertFrame, "ACHIEVEMENT_EARNED", "FirstLootEver_"..id)
end

function addon:DisplayOrQueueAlert(id)
	print("First looted: "..(select(2,GetItemInfo(id))))
end

local funcCache_r = setmetatable({},{__index=function(t,avID)
	local readFunc = function(index)
		local bString = addon.savedVar.achievementMemory[avID] or 0
		return (bit.band(bString,(2^(index-1)))>0)
	end
	t[avID] = readFunc
	return readFunc
end})
local funcCache_w = setmetatable({},{__index=function(t,avID)
	writeFunc = function(index, value)
		if value then
			addon.savedVar.achievementMemory[avID] =  bit.bor(addon.savedVar.achievementMemory[avID] or 0, (2^(index-1)))
		else
			local bString = addon.savedVar.achievementMemory[avID] or 0
			addon.savedVar.achievementMemory[avID] = bString-bit.band(bString,(2^(index-1)))
		end
	end
	t[avID] = writeFunc
	return writeFunc
end})

function addon:CheckAchievementEligibility(id)
	for avId, data in pairs(reverseAchievements) do
		if not self.savedVar.achievements[avId] then
			if data.checkFunc(self, id, funcCache_r[avId], funcCache_w[avId]) then
				self:AwardAchievement(avId)
			end
		end
	end
	AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "CRITERIA_UPDATE")
	AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "TRACKED_ACHIEVEMENT_UPDATE")
	WatchFrame_OnEvent(WatchFrame, "TRACKED_ACHIEVEMENT_UPDATE")
end

local mainFrame = CreateFrame("Frame", "FirstLootEver_Main", UIParent)
mainFrame:SetFrameStrata("DIALOG")
mainFrame:SetSize(480, 544)
mainFrame:SetBackdrop({
	bgFile="Interface\\DialogFrame\\UI-DialogBox-Background",
	edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
	tileSize=32,
	tile=true,
	insets={
		top=12,
		bottom=12,
		right=12,
		left=12,
	},
})
mainFrame.header = CreateFrame("Button", nil, mainFrame)
mainFrame.header:SetNormalTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
mainFrame.header:SetSize(239.4,61.2)
mainFrame.header:SetHitRectInsets(90.72,90.72,12.96,41.76)
mainFrame.header.text = mainFrame.header:CreateFontString()
mainFrame.header.text:SetPoint("TOP",0,-12.16)
mainFrame.header.text:SetFont("Fonts\\FRIZQT__.TTF", 14, "")
mainFrame.header.text:SetTextColor(1,0.82,0)
mainFrame.header.text:SetText("Item lists")
mainFrame.header:SetMovable(true)
mainFrame.header:SetScript("OnMouseDown", function(self)
	self:StartMoving()
end)
mainFrame.header:SetScript("OnMouseUp", function(self)
	self:StopMovingOrSizing()
end)
mainFrame.header:SetPoint("CENTER",UIParent,"CENTER",0,272)
mainFrame:SetPoint("TOP", mainFrame.header, "TOP", 0, -23.28)
--mainFrame:SetScale(1.6)
mainFrame:Hide()

mainFrame.filters = CreateFrame("Frame", nil, mainFrame)
mainFrame.filters:SetPoint("TOPLEFT", 15, -15)
mainFrame.filters:SetPoint("TOPRIGHT", -15, -15)
mainFrame.filters:SetHeight(130)

mainFrame.currentFilter = {[1]="Chest",[2]="Epic",[3]="Cloth",[4]=false}

local function matchFilter(toMatch)
	for key, value in pairs(mainFrame.currentFilter) do
		if (value == false) then
			if toMatch[key] then
				return false
			end
		else
			if not(toMatch[key] == value) then
				return false
			end
		end
	end
	return true
end

mainFrame.filters.entries = setmetatable({}, {__index=function(t,v)
	local f = CreateFrame("Frame", nil, mainFrame.filters)
	if v > 1 then
		f:SetPoint("TOPLEFT", t[v-1], "TOPRIGHT", 5, 0)
		f:SetPoint("BOTTOMLEFT", t[v-1], "BOTTOMRIGHT", 5, 0)
	else
		f:SetPoint("TOPLEFT")
		f:SetPoint("BOTTOMLEFT")
	end
	f.label = f:CreateFontString()
	f.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	f.label:SetTextColor(1,0.82,0)
	f.label:SetPoint("TOP", 0, -2)
	f.label:SetHeight(20)
	f.filters = CreateFrame("ScrollFrame", nil, f)
	f.filters:SetPoint("TOPLEFT", 2, -24)
	f.filters:SetPoint("BOTTOMRIGHT", -2, 2)
	f.filters.child = CreateFrame("Frame", nil, f.filters)
	f.filters.child:EnableMouseWheel(true)
	f.filters.child:SetScript("OnMouseWheel", function(self, delta)
		if delta == 1 then
			f.filters:SetVerticalScroll(math.max(f.filters:GetVerticalScroll()-10,0))
		else
			f.filters:SetVerticalScroll(math.min(f.filters:GetVerticalScroll()+10,f.filters:GetVerticalScrollRange()))
		end
	end)
	f.filters.child:SetWidth(f.filters:GetWidth())
	f.filters:SetScrollChild(f.filters.child)
	f.filters:SetScript("OnSizeChanged", function(self, width)
		self.child:SetWidth(width)
		self:UpdateScrollChildRect()
	end)
	f.filters.child.entries = setmetatable({},{__index=function(iT,iV)
		local iF = CreateFrame("Frame", nil, f.filters.child)
		if iV > 1 then
			iF:SetPoint("TOPLEFT", iT[iV-1], "BOTTOMLEFT", 0, -4)
			iF:SetPoint("TOPRIGHT", iT[iV-1], "BOTTOMRIGHT", 0, -4)
		else
			iF:SetPoint("TOPLEFT")
			iF:SetPoint("TOPRIGHT")
		end
		iF:SetHeight(21)
		iF.check = CreateFrame("CheckButton", nil, iF, "UICheckButtonTemplate")
		iF.check:SetSize(18,18)
		iF.check:SetPoint("LEFT", 4, 0)
		iF.check:SetScript("PostClick", function(self)
			for key, iFr in ipairs(iT) do
				if key ~= iV then
					iFr.check:SetChecked(false)
				end
			end
			self:SetChecked(true)
			mainFrame.currentFilter[f.filterLabel] = iF.optionValue
			mainFrame.list:Update()
		end)
		iF.label = iF:CreateFontString()
		iF.label:SetPoint("LEFT", iF.check, "RIGHT", 4, 0)
		iF.label:SetPoint("RIGHT", -4, 0)
		iF.label:SetWordWrap(false)
		iF.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "")
		iF.label:SetTextColor(1,1,1)
		iF.label:SetJustifyH("LEFT")
		--[[iF.check:SetCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check")
		iF.check:SetNormalTexture("Interface\\Buttons\\--]]
		iT[iV]=iF
		return iF
	end})
	function f:Populate(label, list, filterLabel, defaultValue)
		if not filterLabel then filterLabel = label end
		if self.filterLabel then
			mainFrame.currentFilter[filterLabel] = nil
		end
		self.filterLabel = filterLabel
		self.label:SetText(label)
		for _, btn in ipairs(self.filters.child.entries) do btn:Hide() end
		local c=1
		-- i'm making this not available for the time being. it's far too likely to just bust the client
		self.filters.child.entries[c].label:SetText("Any")
		self.filters.child.entries[c].check:SetChecked(defaultValue == nil)
		self.filters.child.entries[c]:Show()
		c=c+1
		self.filters.child.entries[c].label:SetText("None")
		self.filters.child.entries[c].optionValue = false
		self.filters.child.entries[c].check:SetChecked(defaultValue == false)
		self.filters.child.entries[c]:Show()
		c=c+1
		for _, option in ipairs(list) do
			self.filters.child.entries[c].label:SetText(option)
			self.filters.child.entries[c].optionValue = option
			self.filters.child.entries[c].check:SetChecked(defaultValue==option)
			self.filters.child.entries[c]:Show()
			c=c+1
		end
		self.filters.child:SetHeight((25*(c-1))-4)
		--self.filters:UpdateScrollChildRect()
	end
	t[v] = f
	return f
end})
function mainFrame.filters:Update()
	for _, btn in ipairs(self.entries) do
		btn:Hide()
	end
	local widthEach = (450/#filterOptions)-5
	for num, filterOpt in ipairs(filterOptions) do
		self.entries[num]:Populate(filterLabels[num], filterOpt, num, mainFrame.currentFilter[num])
		self.entries[num]:Show()
		self.entries[num]:SetWidth(widthEach)
	end
end

function mainFrame:Update()
	self.filters:Update()
	self.list:Update()
end
mainFrame:SetScript("OnShow", mainFrame.Update)

mainFrame.list = CreateFrame("ScrollFrame", nil, mainFrame)
mainFrame.list:SetPoint("BOTTOMLEFT", 15, 15)
mainFrame.list:SetPoint("BOTTOMRIGHT", -15, 15)
mainFrame.list:SetHeight(380)
mainFrame.list.child = CreateFrame("Frame", nil, mainFrame.list)
mainFrame.list.child:SetWidth(450)
mainFrame.list.child:EnableMouseWheel(true)
mainFrame.list.child:SetScript("OnMouseWheel", function(self, delta)
	if delta == 1 then
		mainFrame.list:SetVerticalScroll(math.max(mainFrame.list:GetVerticalScroll()-40,0))
	else
		mainFrame.list:SetVerticalScroll(math.min(mainFrame.list:GetVerticalScroll()+40,mainFrame.list:GetVerticalScrollRange()))
	end
end)
mainFrame.list:SetScrollChild(mainFrame.list.child)
mainFrame.list.entries = setmetatable({}, {__index=function(t,v)
	local f = CreateFrame("Frame", nil, mainFrame.list.child)
	f:SetWidth(90)
	f:SetHeight(20)
	f:EnableMouse(true)
	f:SetScript("OnEnter", function(self)
		if self.link then
			GameTooltip:ClearLines()
			GameTooltip:SetOwner(mainFrame, "ANCHOR_NONE")
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("TOPRIGHT", mainFrame, "TOPLEFT", -5, -155)
			GameTooltip:SetHyperlink(self.link)
			if self.obtTime then
				local obtData = date("*t", self.obtTime)
				GameTooltip:AddLine(("Item first looted on %2$d/%1$d/%3$d"):format(obtData.day, obtData.month, obtData.year),0,1,0)
			else
				GameTooltip:AddLine("Item not yet looted",0,1,0)
			end
			GameTooltip:Show()
		end
	end)
	f:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
	f:SetBackdrop({
		bgFile="Interface\\FriendsFrame\\UI-Toast-Background",
		tileSize=12,
		edgeFile="Interface\\FriendsFrame\\UI-Toast-Border",
		tile=1,
		edgeSize=12,
		insets={
			top=5,
			bottom=5,
			right=5,
			left=5,
		},
	})
	f:SetPoint("TOPLEFT", ((v-1)%5)*90, math.floor((v-1)/5)*(-20))
	f.icon = f:CreateTexture()
	f.icon:SetPoint("LEFT", 2, 0)
	f.icon:SetSize(16,16)
	f.label = f:CreateFontString()
	f.label:SetPoint("LEFT", f.icon, "RIGHT", 5, 0)
	f.label:SetPoint("RIGHT", -2, 0)
	f.label:SetWordWrap(false)
	f.label:SetJustifyH("LEFT")
	f.label:SetFont("Fonts\\FRIZQT__.TTF", 10, "OUTLINE")
	function f:Set(itemID, hasItem)
		local name, link, quality, _, _, _, _, _, _, icon = GetItemInfo(itemID)
		if not name then
			self.link = nil
			self:Show()
			self:SetBackdropColor(0.3,0.3,0.3)
			self:SetBackdropBorderColor(0.3,0.3,0.3)
			self.label:SetTextColor(0.3,0.3,0.3)
			self.icon:SetTexture(nil)
			self.label:SetText("(#"..itemID..")")
			return
		end
		local r,g,b=GetItemQualityColor(quality)
		self.obtTime = hasItem
		self.link = link
		self.label:SetText(name)
		self.icon:SetTexture(icon)
		if hasItem then
			self:SetBackdropColor(r,g,b)
			self:SetBackdropBorderColor(r,g,b)
			self.label:SetTextColor(r,g,b)
			SetDesaturation(self.icon, false)
		else
			self:SetBackdropColor(r*0.5,g*0.5,b*0.5)
			self:SetBackdropBorderColor(r*0.5,g*0.5,b*0.5)
			self.label:SetTextColor(r*0.5,g*0.5,b*0.5)
			SetDesaturation(self.icon, true)
		end
		self:Show()
	end
	t[v] = f
	return f
end})

function mainFrame.list:Update()
	local entryNum=1
	local hasItems, noHasItems = {}, {}
	-- loop through all items
	for itemID, itemFilters in pairs(itemList) do
		if matchFilter(itemFilters) then
			local obtainTime = addon.savedVar.items[itemID]
			if obtainTime then
				hasItems[itemID] = obtainTime
			else
				tinsert(noHasItems, itemID)
			end
		end
	end
	for id, obtTime in pairs(hasItems) do
		self.entries[entryNum]:Set(id, obtTime)
		entryNum=entryNum+1
	end
	for _, id in ipairs(noHasItems) do
		self.entries[entryNum]:Set(id, false)
		entryNum=entryNum+1
	end
	for entCnt=entryNum, #self.entries do
		self.entries[entCnt]:Hide()
	end
	self.child:SetHeight(math.ceil((entryNum-1)/5)*20)
	self:UpdateScrollChildRect()
	self:SetVerticalScroll(0)
end

addon:SetScript("OnEvent", function(self, event, ...)
	if event == "ADDON_LOADED" and (...) == "FirstLootEver" then
		self:UnregisterEvent("ADDON_LOADED")
		FirstLootEver_Items = FirstLootEver_Items or {}
		local r = GetRealmName()
		FirstLootEver_Items[r] = FirstLootEver_Items[r] or {achievements={},items={},achievementMemory={},trackedAchievements={}}
		self.savedVar = FirstLootEver_Items[r]
		if not self.savedVar.items then
			-- this is a sVar from an old version, as a new one would already have the items table created
			local a = self.savedVar.achievements
			local i = {}
			for itemID, t in pairs(self.savedVar) do
				if itemID ~= "achievements" then
					i[itemID] = t
				end
			end
			table.wipe(self.savedVar)
			self.savedVar.achievements = a
			self.savedVar.items = i
			self.savedVar.trackedAchievements = {}
		end
		if not self.savedVar.achievementMemory then
			self.savedVar.achievementMemory = {}
		end
		self.isLoaded = true
		if self.scheduledGenerate then
			self.scheduledGenerate = false
			self:RegenerateAchievementCategoryListings()
		end
		if self.scheduledTrackingUpdate then
			self.scheduledTrackingUpdate = false
			AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "TRACKED_ACHIEVEMENT_UPDATE")
			WatchFrame_OnEvent(WatchFrame, "TRACKED_ACHIEVEMENT_UPDATE")
		end
		dummyFunc()
	end
	if event == "UPDATE_INSTANCE_INFO" then
		-- i randomly picked an event that is guaranteed to fire AFTER the initial BAG_UPDATE spam - just to make sure achievements aren't swallowed by the loading screen
		self:UnregisterEvent("UPDATE_INSTANCE_INFO")
		self:RegisterEvent("BAG_UPDATE")
	end
	if event == "BAG_UPDATE" then
		for bag=BACKPACK_CONTAINER, NUM_BAG_SLOTS do
			for slot=1, GetContainerNumSlots(bag) do
				local id = GetContainerItemID(bag,slot)
				if id then -- not an empty slot
					if not self.savedVar.items[id] then
						self.savedVar.items[id] = time()
						addon:DisplayOrQueueAlert(id)
						addon:CheckAchievementEligibility(id)
					end
				end
			end
		end
	end
end)
addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("UPDATE_INSTANCE_INFO")

LoadAddOn("Blizzard_AchievementUI")
hooksecurefunc("AchievementFrameCategories_GetCategoryList", function(t)
	if AchievementFrame.selectedTab == 1 then
		tinsert(t, (#t), {id=FIRST_CATEGORY})
		for blizzID=(FIRST_CATEGORY+1), (CURRENT_CATEGORY-1) do
			tinsert(t, (#t), {id=blizzID, parent=FIRST_CATEGORY, hidden=true})
		end
	end
end)
if AchievementFrame.selectedTab == 1 then
	tinsert(ACHIEVEMENTUI_CATEGORIES,(#ACHIEVEMENTUI_CATEGORIES),{id=FIRST_CATEGORY})
end



-- pre-hook achievement functions to display our category and its contents
local old_GetCategoryInfo = GetCategoryInfo
function GetCategoryInfo(id)
	if id == FIRST_CATEGORY then
		return "FirstLootEver", -1, 0
	elseif reverseBlizzAchievements[id] then
		return reverseBlizzAchievements[id], FIRST_CATEGORY, 0
	else
		return old_GetCategoryInfo(id)
	end
end

local old_GetCategoryNumAchievements = GetCategoryNumAchievements
function GetCategoryNumAchievements(...)
	local catName
	if (...) == FIRST_CATEGORY then
		catName = ""
	elseif reverseBlizzAchievements[(...)] then
		catName = reverseBlizzAchievements[(...)]
	end
	if catName then
		local showAll = select(2,...)
		if showAll then
			if not achievementGroups[catName] then return 0,0 end
			local completed, all = 0, 0
			for _, d in pairs(achievementGroups[catName]) do
				all = all+#d
				for _, a in ipairs(d) do
					if addon:HasAchievement(a.id) then
						completed = completed+1
					end
				end
			end
			return all, completed
		else
			if not categoryListingCache[catName] then
				addon:RegenerateAchievementCategoryListings()
			end
			return #categoryListingCache[catName], categoryListingCache[catName].numCompleted
		end
	else
		-- print("Falling through: GetCategoryNumAchievements")
		return old_GetCategoryNumAchievements(...)
	end
end

local old_GetAchievementInfo = GetAchievementInfo
function GetAchievementInfo(...)
	local achievId
	if select('#', ...) == 2 then
		local id = select(1,...)
		local catName
		if id == FIRST_CATEGORY then
			catName = ""
		else
			catName = reverseBlizzAchievements[id]
		end
		if catName then
			if not categoryListingCache[catName] then
				addon:RegenerateAchievementCategoryListings()
			end
			achievId = categoryListingCache[catName][(select(2,...))]
			if not achievId then
				print("Achievement #"..(select(2,...)).." could not be found in category "..id)
				-- prevent the blizz achievement UI from breaking
				return -1, "Invalid achievement", 0, false, nil, nil, nil, "Invalid achievement", 0, "Interface\\Icons\\Inv_Misc_QuestionMark", "", false
			end
		end
	else
		local id = (...)
		if type(id) == "string" then
			local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
			if ownID then
				achievId = ownID
			end
		end
	end
	if achievId then
		local achievData = reverseAchievements[achievId]
		if not achievData then
			print("Achievement ID "..achievId.." could not be found")
			-- prevent the blizz achievement UI from breaking...again
			return -1, "Invalid achievement", 0, false, nil, nil, nil, "Invalid achievement", 0, "Interface\\Icons\\Inv_Misc_QuestionMark", "", false
		end
		local completionDate = addon.savedVar.achievements[achievId]
		if completionDate then
			local dateInfo = date("*t", completionDate)
			return "FirstLootEver_"..achievId, achievData.name, 0, true, dateInfo.month, dateInfo.day, (dateInfo.year%100), achievData.desc, 0, achievData.icon, "", false
		else
			return "FirstLootEver_"..achievId, achievData.name, 0, false, nil, nil, nil, achievData.desc, 0, achievData.icon, "", false
		end
	else
		return old_GetAchievementInfo(...)
	end
end

local old_IsTrackedAchievement = IsTrackedAchievement
function IsTrackedAchievement(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			return addon.savedVar.trackedAchievements[ownID]
		end
	end
	return old_IsTrackedAchievement(id)
end

local old_AddTrackedAchievement = AddTrackedAchievement
function AddTrackedAchievement(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			addon.savedVar.trackedAchievements[ownID] = true
			AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "TRACKED_ACHIEVEMENT_UPDATE")
			WatchFrame_OnEvent(WatchFrame, "TRACKED_ACHIEVEMENT_UPDATE")
			return true
		end
	end
	return old_AddTrackedAchievement(id)
end

local old_RemoveTrackedAchievement = RemoveTrackedAchievement
function RemoveTrackedAchievement(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			addon.savedVar.trackedAchievements[ownID] = nil
			AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "TRACKED_ACHIEVEMENT_UPDATE")
			WatchFrame_OnEvent(WatchFrame, "TRACKED_ACHIEVEMENT_UPDATE")
			return true
		end
	end
	return old_RemoveTrackedAchievement(id)
end

local old_GetAchievementNumCriteria = GetAchievementNumCriteria
function GetAchievementNumCriteria(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			return (#reverseAchievements[ownID].criteria)
		end
	end
	return old_GetAchievementNumCriteria(id)
end

local old_GetAchievementCriteriaInfo = GetAchievementCriteriaInfo
function GetAchievementCriteriaInfo(...)
	if select('#', ...) == 2 then
		local avID, index = select(1,...)
		if type(avID) == "string" then
			local ownID = tonumber(avID:match("FirstLootEver_(%d+)"))
			if ownID then
				return reverseAchievements[ownID].criteria[index](addon, funcCache_r[avID], funcCache_w[avID])
			end
		end
	end
	return old_GetAchievementCriteriaInfo(...)
end

local old_GetAchievementCategory = GetAchievementCategory
function GetAchievementCategory(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			return blizzAchievements[reverseAchievements[ownID].category]
		end
	end
	return old_GetAchievementCategory(id)
end

local old_GetPreviousAchievement = GetPreviousAchievement
function GetPreviousAchievement(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			return (reverseAchievements[ownID].prevAch and ("FirstLootEver_"..reverseAchievements[ownID].prevAch))
		end
	end
	return old_GetPreviousAchievement(id)
end

local old_GetNextAchievement = GetNextAchievement
function GetNextAchievement(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			return (reverseAchievements[ownID].nextAch and ("FirstLootEver_"..reverseAchievements[ownID].nextAch))
		end
	end
	return old_GetNextAchievement(id)
end

local old_GetAchievementLink = GetAchievementLink
function GetAchievementLink(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			local completed = addon.savedVar.achievements[ownID]
			local flag=0
			for id, func in ipairs(reverseAchievements[ownID].criteria) do
				if (select(3, func(addon))) then
					flag=flag+(2^(id-1))
				end
			end
			if completed then
				local d = date("*t", completed)
				return "|cffffff00|Hachievement:FirstLootEver_"..ownID..":"..UnitGUID("player"):sub(3)..":1:"..d.month..":"..d.day..":"..(d.year%100)..":"..flag..":0:0|h["..reverseAchievements[ownID].name.."]|h|r"
			else
				return "|cffffff00|Hachievement:FirstLootEver_"..ownID..":"..UnitGUID("player"):sub(3)..":0:0:0:-1:"..flag..":0:0|h["..reverseAchievements[ownID].name.."]|h|r"
			end
		end
	end
	return old_GetAchievementLink(id)
end

local old_GetTrackedAchievements = GetTrackedAchievements
function GetTrackedAchievements(...)
	local trackedAchievements = {old_GetTrackedAchievements(...)}
	if addon.isLoaded then
		for ownID, doTrack in pairs(addon.savedVar.trackedAchievements) do
			if doTrack then
				tinsert(trackedAchievements,"FirstLootEver_"..ownID)
			end
		end
	else
		addon.scheduledTrackingUpdate = true
	end
	return unpack(trackedAchievements)
end

local old_GetNumTrackedAchievements = GetNumTrackedAchievements
function GetNumTrackedAchievements(...)
	local i=old_GetNumTrackedAchievements(...)
	for avID, isTracked in pairs(addon.savedVar.trackedAchievements) do
		if isTracked then
			i=i+1
		end
	end
	return i
end


hooksecurefunc(ItemRefTooltip, "SetHyperlink", function(self, link)
	local ownID, unitGUID, completed, month, day, year, flags = link:match("achievement:FirstLootEver_(%d+):([0-9A-Fa-f]+):([0-1]):(%d+):(%d+):([%-0-9]+):(%d+):0:0")
	if ownID then
		local complete = (completed == "1")
		local achievData = reverseAchievements[tonumber(ownID)]
		if achievData then
			self:ClearLines()
			self:AddLine(achievData.name,1,1,1,false)
			self:AddLine(" ") -- empty line
			local playerName = (select(6,GetPlayerInfoByGUID(unitGUID)))
			if complete then
				self:AddLine(ACHIEVEMENT_TOOLTIP_COMPLETE:format(playerName or "Unknown", day, month, year),0,1,0)
			else
				self:AddLine(ACHIEVEMENT_TOOLTIP_IN_PROGRESS:format(playerName or "Unknown"), 0,1,0)
			end
			self:AddLine(" ")
			self:AddLine(achievData.desc, 1,1,1,true)
			self:AddLine(" ")
			flags = tonumber(flags)
			for critCnt=1, #achievData.criteria, 2 do
				local label1 = achievData.criteria[critCnt](addon)
				local has1 = (bit.band(flags,(2^(critCnt-1)))>0)
				local label2, has2
				if #achievData.criteria >= (critCnt+1) then
					label2 = achievData.criteria[critCnt+1](addon)
					has2 = (bit.band(flags,(2^critCnt))>0)
				else
					label2 = " "
					has2 = false
				end
				self:AddDoubleLine(label1, label2, has1 and 0.25 or 0.5, has1 and 0.75 or 0.5, has1 and 0.25 or 0.5, has2 and 0.25 or 0.5, has2 and 0.75 or 0.5, has2 and 0.25 or 0.5)
			end
			self:Show()
		end
	end
end)

SlashCmdList["FIRSTLOOTEVER"] = function() mainFrame[mainFrame:IsShown() and "Hide" or "Show"](mainFrame) end
SLASH_FIRSTLOOTEVER1 = "/fle"

--@alpha@
dummyFunc = function() if not addon:HasAchievement(1339) then addon:AwardAchievement(1339) end end
--@end-alpha@

